#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# @brief NACA64A210 airfoil configuration file for flow trim script
# @authors Adrien Crovato

def getParam():
    p = {}
    # Input/Output
    p['File'] = '../models/n64A410.geo' # Input file containing the model
    p['Pars'] = {'xLength' : 5, 'yLength' : 5, 'sEleFar' : 1.0, 'sEleAirfTE' : 0.01, 'sEleAirfLE' : 0.01} # Parameters for input file model
    p['Dim'] = 2 # Problem dimension
    p['Format'] = 'vtk' # Save format (vtk or gmsh)
    # Markers
    p['Fluid'] = 'field' # Name of physical group containing the fluid
    p['Farfield'] = ['upstream', 'side', 'downstream'] # LIST of name of physical groups containing the farfield boundaries (upstream/downstream should be first/last element)
    p['Wing'] = 'airfoil' # Name of physical group containing the airfoil boundary
    p['Wake'] = 'wake' # Name of physical group containing the wake
    p['Te'] =  'te' # Name of physical group containing the trailing edge
    # Freestream
    p['M_inf'] = 0.65 # Freestream Mach number
    p['CL'] = 0.8 # Target lift coefficient
    p['AoA'] = 3. # Freestream angle of attack (guess)
    p['dCL'] = 6.2 # Slope of CL w.r.t AoA [1/rad] (guess)
    # Geometry
    p['S_ref'] = 1. # Reference surface length (c_ref for 2D)
    p['c_ref'] = 1. # Reference chord length
    p['x_ref'] = 0.25 # Reference point for moment computation (x)
    p['y_ref'] = 0. # Reference point for moment computation (y)
    p['z_ref'] = 0. # Reference point for moment computation (z)
    # Numerical
    p['LSolver'] = 'Pardiso' # Linear solver (Pardiso, GMRES, MUMPS or SparseLU)
    p['NSolver'] = 'Newton' # Noninear solver type (Picard or Newton)
    p['Rel_tol'] = 1e-6 # Relative tolerance on solver residual
    p['Abs_tol'] = 1e-8 # Absolute tolerance on solver residual
    p['Max_it'] = 25 # Solver maximum number of iterations
    p['LS_tol'] = 1e-6 # Tolerance on linesearch residual
    p['Max_it_LS'] = 10 # Linesearch maximum number of iterations
    p['AV_thrsh'] = 1e-2 # Residual threshold below which the artificial viscosity is decreased
    p['M_crit'] = 5. # Critical Mach number above which density is damped
    return p

def main():
    # Script call
    import flow.scripts.trim as t
    trim = t.Trim(getParam())
    trim.run()
    trim.disp()

if __name__ == "__main__":
    main()
