/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wWakeElement.h"
#include "wNode.h"
#include "wMem.h"
#include "wCache.h"
#include "wGauss.h"
using namespace tbox;
using namespace flow;

WakeElement::WakeElement(size_t _no, Element *&_surUpE, Element *&_surLwE,
                         Element *&_volUpE, Element *&_volLwE,
                         std::vector<std::pair<size_t, Node *>> &_wkN) : wObject(), no(_no),
                                                                         surUpE(_surUpE), surLwE(_surLwE), volUpE(_volUpE), volLwE(_volLwE),
                                                                         wkN(_wkN)
{
    // Sanity checks
    if (surUpE->type() != surLwE->type())
        throw std::runtime_error("WakeElement: upper and lower surface element are not the same type!\n");
    if (surUpE->type() == ELTYPE::LINE2)
        nDim = 2;
    else if (surUpE->type() == ELTYPE::TRI3)
        nDim = 3;
    else
    {
        std::stringstream err;
        err << "WakeElement only implemented for surface elements of type "
            << ELTYPE::LINE2 << " or " << ELTYPE::TRI3
            << " (" << surUpE->type() << " was given)!\n";
        throw std::runtime_error(err.str());
    }
    if (volUpE->type() == ELTYPE::TRI3 || volUpE->type() == ELTYPE::TETRA4)
        nColUp = volUpE->nodes.size();
    else
    {
        std::stringstream err;
        err << "WakeElement only implemented for volume elements of type "
            << ELTYPE::TRI3 << " or " << ELTYPE::TETRA4
            << " (" << volUpE->type() << " was given)!\n";
        throw std::runtime_error(err.str());
    }
    if (volLwE->type() == ELTYPE::TRI3 || volLwE->type() == ELTYPE::TETRA4)
        nColLw = volLwE->nodes.size();
    else
    {
        std::stringstream err;
        err << "WakeElement only implemented for volume elements of type "
            << ELTYPE::TRI3 << " or " << ELTYPE::TETRA4
            << " (" << volLwE->type() << " was given)!\n";
        throw std::runtime_error(err.str());
    }
    // Map the nodes
    mapNodes();
    nRow = wkN.size();
}

/**
 * @brief Map the wake nodes to their local volume node indices
 * @authors Adrien Crovato
 */
void WakeElement::mapNodes()
{
    // Find local volume index of contributing surface nodes
    for (auto p : wkN)
    {
        for (size_t i = 0; i < volUpE->nodes.size(); ++i)
        {
            if (p.second == volUpE->nodes[i])
            {
                vsMap[p.first] = i;
                break;
            }
        }
    }
    // Check that all contributing surface nodes have been mapped
    if (vsMap.size() != wkN.size())
    {
        std::stringstream err;
        err << "flow::WakeElement: could not map some contributing nodes from surface element " << *surUpE << " to " << *volUpE << " (number of mapped nodes: " << vsMap.size() << ")!\n";
        throw std::runtime_error(err.str());
    }
}

/**
 * @brief Compute stiffness matrices associated to Wake
 * 
 * K_ij = int V*dN_j * V_inf*dN_i dS
 *
 *              3
 *              o
 *             / \
 *            /   \
 *           /     \
 *     1,1' o-x---x-o 2,2'
 *            A   B
 *
 *     I = sum_j w'_j psi_j V_inf dN'_j V_j invJ_j dN_j phi_j dtm(J')_j
 *       = psi sum_j (w'_j V_inf dN'_j V_j invJ_j dN_j dtm(J')_j) phi
 *       = (psi)|1X2 (sum_j K_j)|2X3 (phi)|3X1
 *
 *     K = w'[dxiN'1 detN'1]T(inv([dx/dxi dy/dxi]))[Vinf,x][Vx Vy]inv([dx/dxi dy/dxi])[dxiN1 dxiN2 dxiN3] dtm([dx/dxi' dy/dxi']^2)
 *           [dxiN'1 detN'1] (   ([dx/det dy/det]))[Vinf,y]          ([dx/det dy/det])[detN1 detN2 detN3]
 *
 *             phi1  phi2  phi3
 *     K = 1' [ .     .     .  ]
 *         2' [ .     .     .  ]
 *
 *     NB: K should be eval at GP A and B, for upper and lower element
 *     NB: V = grad(phi)
 * @authors Adrien Crovato
 */
void WakeElement::buildK(Eigen::MatrixXd &Kup, Eigen::MatrixXd &Klw, std::vector<double> const &u, double alpha)
{
    // Get shape functions and Gauss points
    Cache &surCache = surUpE->getVCache();
    Gauss &surGauss = surCache.getVGauss();
    Eigen::MatrixXd const &volUpDff = volUpE->getVCache().getDsf(0);
    Eigen::MatrixXd const &volLwDff = volLwE->getVCache().getDsf(0);
    // Get Jacobian
    Mem &surMem = surUpE->getVMem();
    Eigen::MatrixXd const &volUpJ = volUpE->getVMem().getJinv(0);
    Eigen::MatrixXd const &volLwJ = volLwE->getVMem().getJinv(0);
    // Reset matrices
    Kup = Eigen::MatrixXd::Zero(nRow, nColUp);
    Klw = Eigen::MatrixXd::Zero(nRow, nColLw);

    // Volume contribution
    // V_inf & dN' (stabilization)
    Eigen::VectorXd Vi = Eigen::VectorXd::Zero(nDim);
    Vi(0) = cos(alpha);
    Vi(nDim - 1) = sin(alpha);
    Eigen::MatrixXd dN(nRow, nDim);
    for (size_t i = 0; i < nRow; ++i)
        for (size_t j = 0; j < nDim; ++j)
            dN(i, j) = volUpDff(j, vsMap.at(wkN[i].first));
    Eigen::VectorXd JdNV(nRow);
    JdNV = dN * volUpJ.transpose() * Vi;

    // A = V*[inv(J)*dN]
    Eigen::RowVectorXd Aup(nColUp), Alw(nColLw);
    Aup = volUpE->computeGrad(u, 0).transpose() * volUpJ * volUpDff;
    Alw = volLwE->computeGrad(u, 0).transpose() * volLwJ * volLwDff;

    // Build
    for (size_t k = 0; k < surGauss.getN(); ++k)
    {
        // N
        Eigen::VectorXd const &ff = surCache.getSf(k);
        Eigen::VectorXd N(nRow);
        for (size_t i = 0; i < nRow; ++i)
            N(i) = 0.5 * sqrt(surMem.getVol()) * JdNV(i) + ff(wkN[i].first);

        // K = K + N*A*w*dtm
        Kup += N * Aup * surGauss.getW(k) * surMem.getDetJ(k);
        Klw += N * Alw * surGauss.getW(k) * surMem.getDetJ(k);
    }
}

/**
 * @brief Compute stiffness matrices associated to Kutta for Newton-Raphson
 * 
 * K_ij = int 2*V*dN_j * V_inf*dN_i dS
 * @authors Adrien Crovato
 */
void WakeElement::buildNK(Eigen::MatrixXd &Kup, Eigen::MatrixXd &Klw, std::vector<double> const &u, double alpha)
{
    // Build
    buildK(Kup, Klw, u, alpha);
    Kup *= 2.;
    Klw *= 2.;
}

/**
 * @brief Compute flux vectors associated to Kutta for Newton-Raphson
 * 
 * s_i = int V*V * V_inf*dN_i dS
 * @authors Adrien Crovato
 */
void WakeElement::buildNs(Eigen::VectorXd &bUp, Eigen::VectorXd &bLw, std::vector<double> const &u, double alpha)
{
    // Get shape functions and Gauss points
    Cache &surCache = surUpE->getVCache();
    Gauss &surGauss = surCache.getVGauss();
    Eigen::MatrixXd const &volUpDff = volUpE->getVCache().getDsf(0);
    // Get Jacobian
    Mem &surMem = surUpE->getVMem();
    Eigen::MatrixXd const &volUpJ = volUpE->getVMem().getJinv(0);
    // Reset matrices
    bUp = Eigen::VectorXd::Zero(nRow);
    bLw = Eigen::VectorXd::Zero(nRow);

    // Volume contribution
    // V_inf & dN' (stabilization)
    Eigen::VectorXd Vi = Eigen::VectorXd::Zero(nDim);
    Vi(0) = cos(alpha);
    Vi(nDim - 1) = sin(alpha);
    Eigen::MatrixXd dN(nRow, nDim);
    for (size_t i = 0; i < nRow; ++i)
        for (size_t j = 0; j < nDim; ++j)
            dN(i, j) = volUpDff(j, vsMap.at(wkN[i].first));
    Eigen::VectorXd JdNV(nRow);
    JdNV = dN * volUpJ.transpose() * Vi;

    // V2
    double grad2Up = volUpE->computeGrad2(u, 0);
    double grad2Lw = volLwE->computeGrad2(u, 0);

    // Build
    for (size_t k = 0; k < surGauss.getN(); ++k)
    {
        // N
        Eigen::VectorXd const &ff = surCache.getSf(k);
        Eigen::VectorXd N(nRow);
        for (size_t i = 0; i < nRow; ++i)
            N(i) = 0.5 * sqrt(surMem.getVol()) * JdNV(i) + ff(wkN[i].first);

        // Compute flux vector
        bUp += N * grad2Up * surGauss.getW(k) * surMem.getDetJ(k);
        bLw += N * grad2Lw * surGauss.getW(k) * surMem.getDetJ(k);
    }
}

void WakeElement::write(std::ostream &out) const
{
    out << "WakeElement #" << no
        << "\n\ton surface elements (" << surUpE->type() << "): " << surUpE->no << " and " << surLwE->no
        << "\n\ton volume elements: (" << volUpE->type() << "): " << volUpE->no << " and " << volLwE->no << std::endl;
}