#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# FE² for "heat" - version "MPI task pull"

from __future__ import print_function
from __future__ import division
from builtins import zip
from builtins import next
from builtins import range
from builtins import object
from past.utils import old_div
import heat as h
import tbox
import math

verb = False  # set to True for (a lot of) debug info

try:
    import mpi4py.MPI as mpi
    comm = mpi.COMM_WORLD
    rank = comm.rank
    siz  = comm.size
    name = mpi.Get_processor_name()
    status = mpi.Status()
    print("info: MPI found")
except:
    comm = None
    rank = 0
    siz  = 1
    name = "noname"
    print("info: MPI not found => MPI disabled")

def barrier():
    if comm:
        comm.Barrier()

def enum(*sequential, **named):
    enums = dict(list(zip(sequential, list(range(len(sequential))))), **named)
    return type('Enum', (), enums)

tags = enum('READY', 'DONE', 'EXIT', 'START', 'WAKEUP', 'SUICIDE')



class Job(object):
    """ Class containing data transmitted between MPI procs
    """
    def __init__(self, id, TM, dTxM, dTyM):
        # input
        self.id = id
        self.TM   = TM
        self.dTxM = dTxM
        self.dTyM = dTyM
        # output
        self.K11 = 0.0
        self.K22 = 0.0 # useless (=K11)
        self.K12 = 0.0
        self.K21 = 0.0 # useless (=-K12)
        
    def execute(self, fem):
        """
        [executed by worker processes]
        solve a given job and calculate "k"
        """
        TM   = self.TM
        dTxM = self.dTxM
        dTyM = self.dTyM 
    
        fem.p.reset(TM, dTxM, dTyM)
        fem.solver.start(fem.gmshWriter) 
        
        # resulting flux qM
        qMx = fem.p.qM[0]
        qMy = fem.p.qM[1]     
 
        if verb:
            print("[%d] qMx = %f" % (rank,qMx)) 
            print("[%d] qMy = %f" % (rank,qMy)) 
        
        # convert flux to an anisotropic conductivity "k(2,2)"
        dTnorm = math.sqrt(dTxM*dTxM+dTyM*dTyM)
        qnorm  = math.sqrt(qMx*qMx+qMy*qMy)
        ampl   = old_div(qnorm,dTnorm)
        
        if 1:
            cos=-(dTxM*qMx+dTyM*qMy)/dTnorm/qnorm
            sin=-(dTxM*qMy-dTyM*qMx)/dTnorm/qnorm
        else:
            ps=-(dTxM*qMx+dTyM*qMy)/dTnorm/qnorm
            if ps>1.0: ps=1.0
            if ps<-1.0: ps=-1.0
            
            #print ps
            ang = math.acos(ps)
            #print "ang=", ang, "ampl=", ampl
            #print "qMx=", qMx, "ampl*dT=", dTxM*ampl, "res=", qMx-dTxM*ampl
            cos = math.cos(ang)
            sin = math.sin(ang)
            
        self.K11 = cos*ampl
        self.K21 = sin*ampl
        self.K22 = self.K11   # => K22 inutile!
        self.K12 = -self.K21  # => K12 inutile!


class FE2(tbox.Fct2UdU):
    def __init__(self, femicro):
        """
        creation du micro modele
        """
        tbox.Fct2UdU.__init__(self)

        self.fem = femicro     # micro model (utile uniqt si 1 thread)  
        self.jbz = {}          # list of jobs
        
        if verb: print("[%d] creation of the micro model" % rank)

    def eval(self, elem, npg, u, gradu, k, fake):  # tjs exécuté par rank#0
        """
        [executed by the master process]
        compute "k" from "u" and "gradu" 
        at element #elem and gauss point #npg
        if "fake=True" => create a job to be run later in parallel
        """
        if fake: # on cree un job
            TM   = u
            dTxM = gradu[0]
            dTyM = gradu[1]        
            dTnorm = math.sqrt(dTxM*dTxM+dTyM*dTyM)
            if dTnorm==0.0:
                dTxM = 1.0
                dTyM = 1.0
                dTnorm=math.sqrt(dTxM*dTxM+dTyM*dTyM)
                
            if verb: 
                print("[m] TM   =",TM)
                print("[m] dTxM =",dTxM)
                print("[m] dTyM =",dTyM)
            
            job = self.jbz.get((elem.no,npg))
            if job: # job existe deja, on le modifie
                job.TM   = TM
                job.dTxM = dTxM
                job.dTyM = dTyM            
            else: # on ajoute ce job
                self.jbz[(elem.no,npg)] = Job((elem.no,npg), TM, dTxM, dTyM)
                #print "[m] creating job", (elem.no,npg)
 
        else: # on retourne le resultat du job 
            
            job = self.jbz.get((elem.no,npg))
            if not job:
                print("[m] job", (elem.no,npg), "not found!")
            k.resize(2, 2)
            k[0,0] = job.K11
            k[1,1] = job.K22 
            k[1,0] = job.K21
            k[0,1] = job.K12

    def evalall(self):   # tjs exécuté par rank#0
    
        import sys
        sys.stdout.flush()
        
        if verb: print("[0] looping on %d jobs" % len(self.jbz))
        
        # version 1 proc (master works alone)
        if siz==1:
            #for job in self.jbz.itervalues():
            for id,job in self.jbz.items():
                job.execute(self.fem) 
            return

        # version N procs
        it=self.jbz.__iter__()
        njobs=len(self.jbz)
        slaves=list(range(1,siz))
        
        for s in slaves:
            if verb: print("[0] sending wake-up signal to worker", s)
            comm.send(None, dest=s, tag=tags.WAKEUP)
        
        num_workers = siz-1
        closed_workers = 0
        while closed_workers<num_workers:
            # get a msg from any source
            data = comm.recv(source=mpi.ANY_SOURCE, 
                             tag=mpi.ANY_TAG, status=status)
            source = status.Get_source()
            tag = status.Get_tag()
            if tag == tags.READY:
                if verb: print("[0] worker %d is ready" % source)
                # worker is ready - send a new task
                job = None
                try:
                    jobk = next(it) # get next job
                    job = self.jbz[jobk]
                except:
                    pass
                if job:
                    if verb: print("[0] sending job to %d" % source)
                    comm.send(job, dest=source, tag=tags.START)
                else:
                    if verb: print("[0] job list is empty! sending EXIT to %d" % source)
                    comm.send(None, dest=source, tag=tags.EXIT)
                    # other workers are still calculating...
                    # master should wait for them
            elif tag==tags.DONE:
                if verb: print("[0] worker %d gives me its results" % source)
                # worker has done its job - get the results
                job = self.jbz[data.id]
                job.K11=data.K11
                job.K22=data.K22
                job.K12=data.K12
                job.K21=data.K21
            elif tag==tags.EXIT:
                closed_workers+=1
                if verb: print("[0] worker %d exited (%d worker(s) still running)" % (source, num_workers-closed_workers))
                
        if verb: print("[0] done.")
        
class Master(object):
    """
    MPI Process with rank #0
    """
    def __init__(self, macro):
        self.macro = macro

    def start(self):
        """
        master loop
        """
        global rank
        if rank!=0: 
            raise Exception("this routine should be called with MPI rank=0")
        if verb: print("[%d] starting master" % rank)
        self.macro.solver.start(self.macro.gmshWriter)

        self.killslaves()
        
        # show results if requested
        from fwk.wutils import parseargs
        args = parseargs()
        if not args.nogui:
            print("\n<CLOSE gmsh to quit!>")    
            from heat.viewer import GUI
            GUI().open(self.macro.msh.name)
        
    def killslaves(self):
        slaves=list(range(1,siz))
        for s in slaves:
            if verb: print("[%d] sending suicide job to %d" % (rank,s))
            comm.send(None, dest=s, tag=tags.SUICIDE)


class Worker(object):
    """
    MPI Process with rank #1-#n
    """
    def __init__(self, micro):
        self.micro = micro
        
    def start(self):
        """
        worker loop
        """
        global rank
        if rank==0: 
            raise Exception("this routine should be called with MPI rank!=0")
        
        if verb: print("[%d] starting worker" % rank)
        while True:
            comm.recv(source=0, tag=mpi.ANY_TAG, status=status)
            tag = status.Get_tag()
            if tag==tags.WAKEUP:
                if verb: print("[%d] waking up" % rank)
                while True:
                    comm.send(None, dest=0, tag=tags.READY)
                    job = comm.recv(source=0, tag=mpi.ANY_TAG, status=status)
                    tag = status.Get_tag()
                    if tag==tags.START:
                        if verb: print("[%d] starting job" % rank)
                        job.execute(self.micro)
                        if verb: print("[%d] sending job results" % rank)
                        comm.send(job, dest=0, tag=tags.DONE)
                    elif tag==tags.EXIT:
                        if verb: print("[%d] sending exit confirmation" % rank)
                        comm.send(None, dest=0, tag=tags.EXIT)
                        break
            elif tag==tags.SUICIDE:
                if verb: print("[%d] I'm dying..." % rank)
                if rank==2:
                    print(self.micro.solver.tms)
                break

   
