#!/usr/bin/env python
# -*- coding: latin-1; -*-

from __future__ import print_function
from __future__ import division
from builtins import str
from builtins import range
from past.utils import old_div
from katropton.eigenvalues import *
import os
import fwk

import vtk
import tboxVtk
import tboxVtk.reader as vtkR
import tboxVtk.lineExtractor as vtkC


def Kirsch_equations(x, y, center_x, center_y, a, sigma_inf):

    centered_x = x-center_x
    centered_y = y-center_y

    r = np.abs(centered_x+centered_y*1.0j)
    theta = np.angle(centered_x+centered_y*1.0j) - old_div(np.pi, 2)

    half_sigma_inf = (old_div(sigma_inf, 2))

    ar = old_div(a, r)
    ar_p_2 = np.power(ar, 2)
    ar_p_4 = np.power(ar, 4)

    cos2theta = np.cos(2*theta)
    sin2theta = np.sin(2*theta)

    sigma_rr = half_sigma_inf*(1-ar_p_2)+half_sigma_inf * \
        (1-4*ar_p_2+3*ar_p_4)*cos2theta
    sigma_thetatheta = half_sigma_inf * \
        (1+ar_p_2)-half_sigma_inf*(1+3*ar_p_4)*cos2theta
    tau_rtheta = -half_sigma_inf*(1+2*ar_p_2-3*ar_p_4)*sin2theta

    VM = np.sqrt(np.square(sigma_rr)+np.square(sigma_thetatheta) -
                 sigma_rr*sigma_thetatheta+3*np.square(tau_rtheta))

    return VM


def cut_plate(mshName):
    reader = vtkR.Reader()
    reader.open(mshName)
    cutter = vtkC.Cutter(reader.reader.GetOutputPort())
    c_output = cutter.cut(1, [0., 0., 0.5], [0., 0., 1.], tag_name='volume_id')
    pts, tri, vals = cutter.extract(
        c_output, 3, ['u', 'x', 'y', 'z'], atPoint=True)
    pts_2, tri_2, vals_2 = cutter.extract(
        c_output, 3, ['G', 'von Mises'], atPoint=False)
    vals.update(vals_2)

    return pts, tri, vals


def cut_line(mshName, P1, P2):
    reader = vtkR.Reader()
    reader.open(mshName)
    cutter = vtkC.LineExtractor(reader.reader.GetOutputPort())
    c_output = cutter.cut(1, P1, P2, tag_name='volume_id')
    pts, tri, vals = cutter.extract(
        c_output, ['u', 'x', 'y', 'z'], atPoint=True)
    pts_2, tri_2, vals_2 = cutter.extract(
        c_output, ['G', 'von Mises'], atPoint=False)
    vals.update(vals_2)

    return pts, tri, vals


def extract_quad_msh(mshName):
    reader = vtkR.Reader()
    reader.open(mshName)
    rdata = reader.reader.GetOutput()

    _pts = rdata.GetPoints()
    pts = np.zeros((_pts.GetNumberOfPoints(), 2))
    pts_indices = np.zeros((_pts.GetNumberOfPoints(),), dtype=int)
    for i in range(0, pts.shape[0]):
        if _pts.GetPoint(i)[2] <= 0.01:
            for j in range(0, 2):
                pts[i][j] = _pts.GetPoint(i)[j]
            pts_indices[i] = 1
    n_elems = rdata.GetNumberOfCells()
    quad = []
    for i in range(0, n_elems):
        elem = rdata.GetCell(i)
        for j in range(0, 6):
            face = elem.GetFace(j)
            points_id = np.zeros((4,), dtype=int)
            good_face = True
            for k in range(0, 4):
                points_id[k] = face.GetPointId(k)
                if pts_indices[points_id[k]] != 1:
                    good_face = False
                    break
            if good_face:
                if len(quad) == 0:
                    quad = points_id.reshape((1, 4))
                else:
                    quad = np.append(quad, points_id.reshape((1, 4)), axis=0)
    return pts, quad


def plot_2D_cut_values_at_node(x, y, tri, value, v_min, v_max, n_levels=16):
    import matplotlib.pyplot as plt
    value = value.reshape((len(x),))
    levels = np.linspace(v_min, v_max, n_levels)
    print(levels)

    fig = plt.figure(figsize=(8, 8))
    plt.set_cmap('coolwarm')
    plt.tricontour(x, y, tri, value, levels, linestyles='-',
                   colors='black', linewidths=0.5)
    axtri = plt.tricontourf(x, y, tri, value, levels)
    plt.gca().set_aspect('equal')
    plt.gca().set_axis_off()
    plt.subplots_adjust(top=1, bottom=0, right=1, left=0,
                        hspace=0, wspace=0)
    plt.margins(0, 0)
    plt.gca().xaxis.set_major_locator(plt.NullLocator())
    plt.gca().yaxis.set_major_locator(plt.NullLocator())


def plot_2D_cut_values_at_elem(x, y, tri_1, value, v_min, v_max, n_levels=16, shading=False):
    import matplotlib.pyplot as plt

    value = value.reshape((tri_1.shape[0],))

    levels = np.linspace(v_min, v_max, n_levels)
    print(levels)
    fig = plt.figure(figsize=(8, 8))
    plt.set_cmap('coolwarm')
    if shading:
        plt.tripcolor(x, y, tri_1, value, shading='gouraud')
    else:
        plt.tripcolor(x, y, tri_1, value, levels)
    plt.gca().set_aspect('equal')
    plt.gca().set_axis_off()
    plt.subplots_adjust(top=1, bottom=0, right=1, left=0,
                        hspace=0, wspace=0)
    plt.margins(0, 0)
    plt.gca().xaxis.set_major_locator(plt.NullLocator())
    plt.gca().yaxis.set_major_locator(plt.NullLocator())


def plot_2D_mesh(pts, quad):
    import matplotlib.pyplot as plt
    fig = plt.figure(figsize=(8, 8))
    f = open('mesh_2.tex', 'w')
    print('\begin{tikzpicture}', file=f)
    for i in range(0, quad.shape[0]):
        x = np.zeros((quad.shape[1]+1,))
        y = np.zeros((quad.shape[1]+1,))
        for j in range(0, quad.shape[1]):
            x[j] = pts[quad[i, j], 0]
            y[j] = pts[quad[i, j], 1]
        x[-1] = pts[quad[i, 0], 0]
        y[-1] = pts[quad[i, 0], 1]
        string_1 = '\draw[thick] '
        for j in range(0, quad.shape[1]+1):
            string_1 = string_1 + '(' + str(x[j]) + ',' + str(y[j]) + ')'
            if j == quad.shape[1]:
                string_1 = string_1 + ';'
            else:
                string_1 = string_1 + '--'

        print(string_1, file=f)
        plt.plot(x, y, 'black')
    plt.gca().set_aspect('equal')

    print('\end{tikzpicture}', file=f)
    f.closed


def plot_1D_data_per_elem(pts, elems, vals):
    import matplotlib.pyplot as plt

    plt.figure()
    for i in range(0, elems.shape[0]):
        if i == 0:
            x = np.array([pts[elems[i][0]], pts[elems[i][1]]])
            y = np.array([vals[i], vals[i]])
        else:
            x = np.append(x, [pts[elems[i][0]], pts[elems[i][1]]])
            y = np.append(y, [vals[i], vals[i]])

    plt.plot(x, y, 'r')


def main():
    import matplotlib.pyplot as plt

    file_dir = os.path.dirname(__file__)
    work_dir = os.getcwd()

    display = False

    mshName_extended_1 = file_dir + \
        '/../../../workspace/katropton_CMAME_tests_plate_extended/plate_fused_ast1_s0_0'
    mshName_extended_2 = file_dir + \
        '/../../../workspace/katropton_CMAME_tests_plate_extended_1/plate_fused_ast1_s0_0'
    mshName_mean = file_dir + \
        '/../../../workspace/katropton_CMAME_tests_plate/plate_fused_ast1_s0_0'

    i = 0
    for mshName in [mshName_extended_1, mshName_extended_2, mshName_mean]:
        if i == 2:
            pts, tri, vals = cut_line(mshName, [5., 5., 0.5], [10, 5., 0.5])
        else:
            pts, tri, vals = cut_line(mshName, [5., 5., 0.5], [50, 5., 0.5])
        VM_theory = Kirsch_equations(pts+5, np.zeros(pts.shape), 5, 0, 2, 700)
        plot_1D_data_per_elem(pts, tri, vals['von Mises'])
        plt.plot(pts, VM_theory)
        plt.ylabel('[MPa]')
        plt.xlabel('Pos')
        plt.savefig('fig_'+str(i)+'.png')
        i = i + 1
    # plt.show()


if __name__ == "__main__":
    main()
