from __future__ import print_function
from __future__ import division
from builtins import range
from past.utils import old_div
import vtk
import numpy as np
from numpy import zeros
import matplotlib.pyplot as plt
#from mpl_toolkits.mplot3d import Axes3D
#import matplotlib2tikz
from numpy import linalg as LA


def SaveFigureAsImage(fileName, fig=None, **kwargs):
    ''' Save a Matplotlib figure as an image without borders or frames.
       Args:
            fileName (str): String that ends in .png etc.

            fig (Matplotlib figure instance): figure you want to save as the image
        Keyword Args:
            orig_size (tuple): width, height of the original image used to maintain 
            aspect ratio.
    '''
    fig_size = fig.get_size_inches()
    w, h = fig_size[0], fig_size[1]
    fig.patch.set_alpha(0)
    if 'orig_size' in kwargs:  # Aspect ratio scaling if required
        w, h = kwargs['orig_size']
        w2, h2 = fig_size[0], fig_size[1]
        fig.set_size_inches([(old_div(w2, w)) * w, (old_div(w2, w)) * h])
        fig.set_dpi((old_div(w2, w)) * fig.get_dpi())
    a = fig.gca()
    a.set_frame_on(False)
    a.set_xticks([])
    a.set_yticks([])
    plt.axis('off')
    plt.xlim(0, h)
    plt.ylim(w, 0)
    fig.savefig(fileName, transparent=True, bbox_inches='tight',
                pad_inches=0)


def compute_integrated_F(volume_id,
                         component_id,
                         cut_origin_x,
                         cut_origin_y,
                         cut_origin_z,
                         cut_normal_x,
                         cut_normal_y,
                         cut_normal_z,
                         reader,
                         plot_results=False):

    thresh = vtk.vtkThreshold()
    thresh.ThresholdBetween(volume_id - 0.5, volume_id + 0.5)
    thresh.SetInputConnection(reader.GetOutputPort())
    thresh.SetInputArrayToProcess(0, 0, 0,
                                  vtk.vtkDataObject.FIELD_ASSOCIATION_CELLS,
                                  "volume_id")
    thresh.Update()

    plane = vtk.vtkPlane()
    plane.SetOrigin(cut_origin_x, cut_origin_y, cut_origin_z)
    plane.SetNormal(cut_normal_x, cut_normal_y, cut_normal_z)

    cutter = vtk.vtkCutter()
    cutter.SetCutFunction(plane)
    cutter.SetInputConnection(thresh.GetOutputPort())
    cutter.Update()

    data = cutter.GetOutput()

    triangles = data.GetPolys().GetData()
    points = data.GetPoints()

    stress = data.GetCellData().GetArray('sigma')

    ntri = old_div(triangles.GetNumberOfTuples(), 4)
    npts = points.GetNumberOfPoints()
    nstress = stress.GetNumberOfTuples()

    tri = zeros((ntri, 3), dtype=int)
    x = zeros(npts)
    y = zeros(npts)
    z = zeros(npts)

    sigma_vector = zeros(nstress)

    sigma_vector_node = zeros(npts)
    area_node = zeros(npts)

    for i in range(0, ntri):
        tri[i, 0] = triangles.GetTuple(4 * i + 1)[0]
        tri[i, 1] = triangles.GetTuple(4 * i + 2)[0]
        tri[i, 2] = triangles.GetTuple(4 * i + 3)[0]

    for i in range(npts):
        pt = points.GetPoint(i)
        x[i] = pt[0]
        y[i] = pt[1]
        z[i] = pt[2]

    for i in range(0, nstress):
        sigma = stress.GetTuple(i)
        sigma_vector[i] = sigma[component_id]

    if plot_results:
        fig = plt.figure(figsize=(8, 8))
        plt.set_cmap('coolwarm')
        #plt.triplot(x, y, tri)
        axtri = plt.tripcolor(x, y, tri, sigma_vector)
        #plt.scatter(x, y)
        plt.gca().set_aspect('equal')
        cb = fig.colorbar(axtri, orientation="horizontal")

    F = 0.
    total_area = 0.
    for i in range(0, ntri):
        cross = np.cross(
            np.array([
                x[tri[i, 1]] - x[tri[i, 0]], y[tri[i, 1]] - y[tri[i, 0]],
                z[tri[i, 1]] - z[tri[i, 0]]
            ]),
            np.array([
                x[tri[i, 2]] - x[tri[i, 0]], y[tri[i, 2]] - y[tri[i, 0]],
                z[tri[i, 2]] - z[tri[i, 0]]
            ]))
        area = old_div(LA.norm(cross), 2)
        F = F + area * sigma_vector[i]
        total_area = total_area + area
        for j in range(0, 3):
            sigma_vector_node[tri[i, j]] = sigma_vector_node[
                tri[i, j]] + area * sigma_vector[i]
            area_node[tri[i, j]] = area_node[tri[i, j]] + area

    for i in range(npts):
        sigma_vector_node[i] = old_div(sigma_vector_node[i], area_node[i])

    if plot_results:
        fig = plt.figure(figsize=(8, 8))
        plt.set_cmap('coolwarm')
        #plt.triplot(x, y, tri)
        axtri = plt.tricontourf(x, y, tri, sigma_vector_node)
        #plt.scatter(x, y)
        plt.gca().set_aspect('equal')
        cb = fig.colorbar(axtri, orientation="horizontal")

    return F


def post_process():
    filename = 'MIRROR_1_ASSEMBLY_fused_2_ast1_s0_0.vtu'

    reader = vtk.vtkXMLUnstructuredGridReader()
    reader.SetFileName(filename)
    reader.Update()

    #data1 = reader.GetOutput()

    thresh = vtk.vtkThreshold()
    thresh.ThresholdBetween(8.5, 11.5)
    thresh.SetInputConnection(reader.GetOutputPort())
    thresh.SetInputArrayToProcess(0, 0, 0,
                                  vtk.vtkDataObject.FIELD_ASSOCIATION_CELLS,
                                  "volume_id")
    thresh.Update()

    plane = vtk.vtkPlane()
    plane.SetOrigin(-0.0156128890812397, 0.020816631615161896,
                    -0.009049303388172896)
    plane.SetNormal(0, 0, 1)

    cutter = vtk.vtkFiltersCorePython.vtkCutter()
    cutter.SetCutFunction(plane)
    cutter.SetInputConnection(thresh.GetOutputPort())
    cutter.Update()

    data = cutter.GetOutput()

    triangles = data.GetPolys().GetData()
    points = data.GetPoints()

    mapper = vtk.vtkCellDataToPointData()
    mapper.AddInputData(data)
    mapper.Update()

    vels = mapper.GetOutput().GetPointData().GetArray('z')  # .GetArray(1)

    stress = data.GetCellData().GetArray('sigma')
    stress_VM = data.GetCellData().GetArray('von Mises')

    ntri = old_div(triangles.GetNumberOfTuples(), 4)
    npts = points.GetNumberOfPoints()
    nvls = points.GetNumberOfPoints()  # vels.GetNumberOfTuples()

    nstress = stress.GetNumberOfTuples()

    tri = zeros((ntri, 3), dtype=int)
    x = zeros(npts)
    y = zeros(npts)
    z = zeros(npts)
    ux = zeros(nvls)

    sigma_vector_node = zeros(npts)
    area_node = zeros(npts)

    sigma_zz = zeros(nstress)

    for i in range(0, ntri):
        tri[i, 0] = triangles.GetTuple(4 * i + 1)[0]
        tri[i, 1] = triangles.GetTuple(4 * i + 2)[0]
        tri[i, 2] = triangles.GetTuple(4 * i + 3)[0]

    for i in range(npts):
        pt = points.GetPoint(i)
        x[i] = pt[0]
        y[i] = pt[1]
        z[i] = pt[2]

    for i in range(0, nvls):
        U = vels.GetTuple(i)
        ux[i] = U[0]

    for i in range(0, nstress):
        sigma = stress.GetTuple(i)
        sigma_zz[i] = sigma[8]

    for i in range(0, ntri):
        cross = np.cross(
            np.array([
                x[tri[i, 1]] - x[tri[i, 0]], y[tri[i, 1]] - y[tri[i, 0]],
                z[tri[i, 1]] - z[tri[i, 0]]
            ]),
            np.array([
                x[tri[i, 2]] - x[tri[i, 0]], y[tri[i, 2]] - y[tri[i, 0]],
                z[tri[i, 2]] - z[tri[i, 0]]
            ]))
        area = old_div(LA.norm(cross), 2)
        for j in range(0, 3):
            sigma_vector_node[
                tri[i, j]] = sigma_vector_node[tri[i, j]] + area * sigma_zz[i]
            area_node[tri[i, j]] = area_node[tri[i, j]] + area

    for i in range(npts):
        sigma_vector_node[i] = old_div(sigma_vector_node[i], area_node[i])

    #tmp = stress.GetTuple(0)

    levels = np.linspace(min(ux), max(ux), 16)
    # Velocity x-component
    fig = plt.figure(figsize=(8, 8))
    plt.set_cmap('coolwarm')
    plt.tricontour(x,
                   y,
                   tri,
                   ux,
                   levels,
                   linestyles='-',
                   colors='black',
                   linewidths=0.5)
    axtri = plt.tricontourf(x, y, tri, ux, levels)
    plt.gca().set_aspect('equal')

    # plt.minorticks_on()
    # plt.gca().set_xticklabels([])
    # plt.gca().set_yticklabels([])
    # plt.gca().set_axis_off()
    #plt.subplots_adjust(top = 1, bottom = 0, right = 1, left = 0, hspace = 0, wspace = 0)
    # plt.margins(0,0)
    # plt.gca().xaxis.set_major_locator(plt.NullLocator())
    # plt.gca().yaxis.set_major_locator(plt.NullLocator())
    # plt.axis('off')
    #plt.tricontour(x, y, tri, ux, 16)
    # plt.show
    #plt.savefig("Ux.pdf", transparent=True, bbox_inches='tight', pad_inches=-0.05)

    # matplotlib2tikz.save('test.tex')

    cb = fig.colorbar(axtri, orientation="horizontal")

    levels = np.linspace(min(sigma_zz), max(sigma_zz), 16)
    fig = plt.figure(figsize=(8, 8))
    plt.set_cmap('coolwarm')
    axtri = plt.tripcolor(x, y, tri, sigma_zz, levels)
    print("sigma_zz_unsmooth.pdf")
    print(levels)
    plt.gca().set_aspect('equal')
    plt.gca().set_axis_off()
    plt.subplots_adjust(top=1, bottom=0, right=1, left=0, hspace=0, wspace=0)
    plt.margins(0, 0)
    plt.gca().xaxis.set_major_locator(plt.NullLocator())
    plt.gca().yaxis.set_major_locator(plt.NullLocator())
    plt.savefig("sigma_zz_unsmooth.pdf",
                transparent=True,
                bbox_inches='tight',
                pad_inches=-0.05)

    #cb = fig.colorbar(axtri, orientation="horizontal")

    levels = np.linspace(min(sigma_vector_node), max(sigma_vector_node), 16)
    fig = plt.figure(figsize=(8, 8))
    plt.set_cmap('coolwarm')
    axtri = plt.tricontourf(x, y, tri, sigma_vector_node, levels)
    print("sigma_zz_smooth.pdf")
    print(levels)
    plt.gca().set_aspect('equal')
    plt.gca().set_axis_off()
    plt.subplots_adjust(top=1, bottom=0, right=1, left=0, hspace=0, wspace=0)
    plt.margins(0, 0)
    plt.gca().xaxis.set_major_locator(plt.NullLocator())
    plt.gca().yaxis.set_major_locator(plt.NullLocator())
    plt.savefig("sigma_zz_smooth.pdf",
                transparent=True,
                bbox_inches='tight',
                pad_inches=-0.05)

    #cb = fig.colorbar(axtri, orientation="horizontal")

    F_zz_1 = compute_integrated_F(9, 8, 0.0529999991413206, 0.0440000002272427,
                                  -0.0413666241463175, 0, 0, 1, reader)
    F_zz_2 = compute_integrated_F(10, 8, 0.0529999991413206,
                                  0.0440000002272427, -0.0413666241463175, 0,
                                  0, 1, reader)
    F_zz_3 = compute_integrated_F(11, 8, 0.0529999991413206,
                                  0.0440000002272427, -0.0413666241463175, 0,
                                  0, 1, reader)

    print(F_zz_1)
    print(F_zz_2)
    print(F_zz_3)

    # plt.show()
