#include "Graph.h"

#include "wProblem.h"
#include "wMshData.h"
#include "wElement.h"
#include "wNode.h"

#include <utility>

#include <Kokkos_UnorderedMap.hpp>
#include <Kokkos_Sort.hpp>

#include "Tpetra_CrsMatrix.hpp"
#include <MatrixMarket_Tpetra.hpp>

#include "LinearSolver.h"

using namespace katropton;

int min(int a, int b)
{
    return (a <= b) ? a : b;
}

int max(int a, int b)
{
    return (a >= b) ? a : b;
}

/**
 * @brief Graph constructor
 *
 * This constructor computes the sparsity pattern of the primal matrix.
 * 
 * Arguments:
 * <ul>
 *  <li> map: an RCP to a Map object,
 *  <li> elementsList: an RCP to a ElementsList object.
 * </ul>
 */
Graph::Graph(Teuchos::RCP<Map> map, Teuchos::RCP<ElementsList> elementsList)
{
    using std::pair;
    using Teuchos::RCP;
    using Teuchos::rcp;
    using Teuchos::tuple;
    typedef Map::local_ordinal_type local_ordinal_type;
    typedef Map::global_ordinal_type global_ordinal_type;
    typedef Tpetra::CrsGraph<local_ordinal_type, global_ordinal_type> crs_graph_type;
    typedef local_ordinal_type LO;
    typedef global_ordinal_type GO;

    auto numMyElems = elementsList->getElementNumber();

    auto numPrimalDPN = map->numPrimalDPN;

    int numLclNodes = map->mapNodes->getNodeNumElements();

    int numLclRowsK = 3 * numLclNodes;
    int numLclRowsL = numLclNodes;

    Kokkos::View<size_t *> rowCountsK("row counts", numLclRowsK);
    Kokkos::View<size_t *> rowCountsS("row counts", numLclRowsK);
    Kokkos::View<size_t *> rowCountsL("row counts", numLclRowsL);

    const int maxConnections = 30;
    Kokkos::UnorderedMap<pair<GO, GO>, bool> nodenode(numLclNodes * maxConnections);

    size_t numLclEntries = 0;

    Kokkos::parallel_reduce(
        numMyElems,
        KOKKOS_LAMBDA(const LO e, size_t &curNumLclEntries) {
            const int element_size = elementsList->getElementSize(e);

            for (auto n1 = 0; n1 < element_size; ++n1)
                for (auto n2 = n1; n2 < element_size; ++n2)
                {
                    const LO lclNodeWO1 = elementsList->getElementNode(e, n1);
                    const LO lclNodeWO2 = elementsList->getElementNode(e, n2);

                    const GO globalTmpNode1 = map->mapNodesWO->getGlobalElement(lclNodeWO1);
                    const GO globalTmpNode2 = map->mapNodesWO->getGlobalElement(lclNodeWO2);

                    const GO globalNode1 = min(globalTmpNode1, globalTmpNode2);
                    const GO globalNode2 = max(globalTmpNode1, globalTmpNode2);

                    const pair<GO, GO> key(globalNode1, globalNode2);

                    auto result = nodenode.insert(key);

                    if (result.success())
                    {
                        const bool isDiffNodes = (globalNode1 != globalNode2);
                        const bool isLocalNode1 = map->mapNodes->isNodeGlobalElement(globalNode1);
                        const bool isLocalNode2 = map->mapNodes->isNodeGlobalElement(globalNode2);

                        const LO lclNode1 = map->mapNodes->getLocalElement(globalNode1);
                        const LO lclNode2 = map->mapNodes->getLocalElement(globalNode2);

                        if (numPrimalDPN == 4 || numPrimalDPN == 1)
                        {
                            if (isLocalNode1)
                                Kokkos::atomic_fetch_add(&rowCountsL(lclNode1), 1);
                            if (isDiffNodes && isLocalNode2)
                                Kokkos::atomic_fetch_add(&rowCountsL(lclNode2), 1);
                        }
                        if (numPrimalDPN == 4 || numPrimalDPN == 3)
                        {
                            for (auto i = 0; i < 3; ++i)
                                for (auto j = 0; j < 3; ++j)
                                {
                                    const LO lclRow = 3 * lclNode1 + i;
                                    const LO lclCol = 3 * lclNode2 + j;

                                    if (isLocalNode1)
                                        Kokkos::atomic_fetch_add(&rowCountsK(lclRow), 1);
                                    if (isDiffNodes && isLocalNode2)
                                        Kokkos::atomic_fetch_add(&rowCountsK(lclCol), 1);
                                }
                        }
                        if (numPrimalDPN == 4)
                        {
                            if (isLocalNode1)
                                for (auto i = 0; i < 3; ++i)
                                {
                                    const LO lclRow = 3 * lclNode1 + i;
                                    Kokkos::atomic_fetch_add(&rowCountsS(lclRow), 1);
                                }

                            if (isDiffNodes && isLocalNode2)
                                for (auto i = 0; i < 3; ++i)
                                {
                                    const LO lclRow = 3 * lclNode2 + i;

                                    Kokkos::atomic_fetch_add(&rowCountsS(lclRow), 1);
                                }
                        }
                    }
                }
        },
        numLclEntries);

    Kokkos::View<size_t *> rowOffsetsK("row offsets", numLclRowsK + 1);
    Kokkos::View<size_t *> rowOffsetsS("row offsets", numLclRowsK + 1);
    Kokkos::View<size_t *> rowOffsetsL("row offsets", numLclRowsL + 1);

    if (numPrimalDPN == 4 || numPrimalDPN == 1)
        Kokkos::parallel_scan(
            numLclRowsL,
            KOKKOS_LAMBDA(int irow, int &update, bool final) {
                // parallel scan is a multi-pass parallel pattern
                // In the ‘final’ pass ‘update’ has the prefix value
                if (final)
                    rowOffsetsL(irow) = update;
                update += rowCountsL(irow);
                if (final && numLclRowsL == irow + 1)
                    rowOffsetsL(irow + 1) = update; // total non-zeros
            });

    if (numPrimalDPN == 4 || numPrimalDPN == 3)
        Kokkos::parallel_scan(
            numLclRowsK,
            KOKKOS_LAMBDA(int irow, int &update, bool final) {
                // parallel scan is a multi-pass parallel pattern
                // In the ‘final’ pass ‘update’ has the prefix value
                if (final)
                    rowOffsetsK(irow) = update;
                update += rowCountsK(irow);
                if (final && numLclRowsK == irow + 1)
                    rowOffsetsK(irow + 1) = update; // total non-zeros
            });

    if (numPrimalDPN == 4)
        Kokkos::parallel_scan(
            numLclRowsK,
            KOKKOS_LAMBDA(int irow, int &update, bool final) {
                // parallel scan is a multi-pass parallel pattern
                // In the ‘final’ pass ‘update’ has the prefix value
                if (final)
                    rowOffsetsS(irow) = update;
                update += rowCountsS(irow);
                if (final && numLclRowsK == irow + 1)
                    rowOffsetsS(irow + 1) = update; // total non-zeros
            });

    Kokkos::deep_copy(rowCountsK, static_cast<size_t>(0));
    Kokkos::deep_copy(rowCountsS, static_cast<size_t>(0));
    Kokkos::deep_copy(rowCountsL, static_cast<size_t>(0));

    Kokkos::View<LO *> colIndicesK("column indices", rowOffsetsK(numLclRowsK));
    Kokkos::View<LO *> colIndicesS("column indices", rowOffsetsS(numLclRowsK));
    Kokkos::View<LO *> colIndicesL("column indices", rowOffsetsL(numLclRowsL));

    Kokkos::parallel_for(
        nodenode.capacity(),
        KOKKOS_LAMBDA(int ientry) {
            if (nodenode.valid_at(ientry))
            {
                const pair<GO, GO> key = nodenode.key_at(ientry);
                const GO globalNode1 = key.first;
                const GO globalNode2 = key.second;

                const bool isDiffNodes = (globalNode1 != globalNode2);
                const bool isLocalNode1 = map->mapNodes->isNodeGlobalElement(globalNode1);
                const bool isLocalNode2 = map->mapNodes->isNodeGlobalElement(globalNode2);

                const LO lclNode1 = map->mapNodes->getLocalElement(globalNode1);
                const LO lclNode2 = map->mapNodes->getLocalElement(globalNode2);

                const LO lclNodeWO1 = map->mapNodesWO->getLocalElement(globalNode1);
                const LO lclNodeWO2 = map->mapNodesWO->getLocalElement(globalNode2);

                if (numPrimalDPN == 4 || numPrimalDPN == 1)
                {

                    if (isLocalNode1)
                    {
                        const LO lclRow = lclNode1;
                        const LO lclCol = lclNodeWO2;

                        const size_t count = Kokkos::atomic_fetch_add(&rowCountsL(lclRow), 1);
                        colIndicesL(rowOffsetsL(lclRow) + count) = lclCol;
                    }
                    if (isDiffNodes && isLocalNode2)
                    {
                        const LO lclRow = lclNode2;
                        const LO lclCol = lclNodeWO1;

                        const size_t count = Kokkos::atomic_fetch_add(&rowCountsL(lclRow), 1);
                        colIndicesL(rowOffsetsL(lclRow) + count) = lclCol;
                    }
                }
                if (numPrimalDPN == 4 || numPrimalDPN == 3)
                {
                    if (isLocalNode1)
                    {
                        for (auto i = 0; i < 3; ++i)
                            for (auto j = 0; j < 3; ++j)
                            {
                                const LO lclRow = 3 * lclNode1 + i;
                                const LO lclCol = 3 * lclNodeWO2 + j;

                                const size_t count = Kokkos::atomic_fetch_add(&rowCountsK(lclRow), 1);
                                colIndicesK(rowOffsetsK(lclRow) + count) = lclCol;
                            }
                    }
                    if (isDiffNodes && isLocalNode2)
                    {
                        for (auto i = 0; i < 3; ++i)
                            for (auto j = 0; j < 3; ++j)
                            {
                                const LO lclRow = 3 * lclNode2 + i;
                                const LO lclCol = 3 * lclNodeWO1 + j;

                                const size_t count = Kokkos::atomic_fetch_add(&rowCountsK(lclRow), 1);
                                colIndicesK(rowOffsetsK(lclRow) + count) = lclCol;
                            }
                    }
                }
                if (numPrimalDPN == 4)
                {
                    if (isLocalNode1)
                        for (auto i = 0; i < 3; ++i)
                        {
                            const LO lclRow = 3 * lclNode1 + i;
                            const LO lclCol = lclNodeWO2;

                            const size_t count = Kokkos::atomic_fetch_add(&rowCountsS(lclRow), 1);
                            colIndicesS(rowOffsetsS(lclRow) + count) = lclCol;
                        }
                    if (isDiffNodes && isLocalNode2)
                        for (auto i = 0; i < 3; ++i)
                        {
                            const LO lclRow = 3 * lclNode2 + i;
                            const LO lclCol = lclNodeWO1;

                            const size_t count = Kokkos::atomic_fetch_add(&rowCountsS(lclRow), 1);
                            colIndicesS(rowOffsetsS(lclRow) + count) = lclCol;
                        }
                }
            }
        });

    LinearSolver::getTimers()["graph: sort"].start();

    if (numPrimalDPN == 4 || numPrimalDPN == 1)
        Kokkos::parallel_for(
            numLclRowsL,
            KOKKOS_LAMBDA(int lclRow) {
                auto currentColIndices = subview(colIndicesL, Kokkos::make_pair(rowOffsetsL(lclRow), rowOffsetsL(lclRow + 1)));
                LO *const lclColIndsRaw = currentColIndices.data();
                std::sort(lclColIndsRaw, lclColIndsRaw + rowOffsetsL(lclRow + 1) - rowOffsetsL(lclRow));
            });

    if (numPrimalDPN == 4 || numPrimalDPN == 3)
        Kokkos::parallel_for(
            numLclRowsK,
            KOKKOS_LAMBDA(int lclRow) {
                auto currentColIndices = subview(colIndicesK, Kokkos::make_pair(rowOffsetsK(lclRow), rowOffsetsK(lclRow + 1)));
                LO *const lclColIndsRaw = currentColIndices.data();
                std::sort(lclColIndsRaw, lclColIndsRaw + rowOffsetsK(lclRow + 1) - rowOffsetsK(lclRow));
            });

    if (numPrimalDPN == 4)
        Kokkos::parallel_for(
            numLclRowsK,
            KOKKOS_LAMBDA(int lclRow) {
                auto currentColIndices = subview(colIndicesS, Kokkos::make_pair(rowOffsetsS(lclRow), rowOffsetsS(lclRow + 1)));
                LO *const lclColIndsRaw = currentColIndices.data();
                std::sort(lclColIndsRaw, lclColIndsRaw + rowOffsetsS(lclRow + 1) - rowOffsetsS(lclRow));
            });

    LinearSolver::getTimers()["graph: sort"].stop();

    if (numPrimalDPN == 4 || numPrimalDPN == 1)
    {
        L = rcp(new crs_graph_type(map->thermoMap, map->thermoMapWO, rowOffsetsL, colIndicesL));
        L->fillComplete(map->thermoMap, map->thermoMap);
    }

    if (numPrimalDPN == 4 || numPrimalDPN == 3)
    {
        K = rcp(new crs_graph_type(map->mechaMap, map->mechaMapWO, rowOffsetsK, colIndicesK));
        K->fillComplete(map->mechaMap, map->mechaMap);
    }

    if (numPrimalDPN == 4)
    {
        S = rcp(new crs_graph_type(map->mechaMap, map->thermoMapWO, rowOffsetsS, colIndicesS));
        S->fillComplete(map->thermoMap, map->mechaMap);

        Tpetra::MatrixMarket::Writer<Tpetra::CrsMatrix<>>::writeSparseGraphFile(std::string("S_graph.txt"), S);
    }
}
