#ifndef KATROPTON_MATRICES_H
#define KATROPTON_MATRICES_H

#include "katropton.h"

#include "Map.h"
#include "Graph.h"

#include <Tpetra_Map.hpp>
#include <Tpetra_Vector.hpp>
#include <Tpetra_CrsGraph.hpp>
#include <Tpetra_CrsMatrix.hpp>
#include <Teuchos_RCP.hpp>

#include <Xpetra_BlockedMap.hpp>
#include <Xpetra_BlockedCrsMatrix.hpp>
#include <Stokhos_MueLu_MP_Vector.hpp>

namespace katropton
{

/**
 * @brief Class which includes all the Trilinos matrices (Tpetra matrices and Xpetra matrices) used
 * in the simulation.
 * 
 * In the case of the saddle point matrix arising from the Mortar FEM, the block matrix is stored as follows:
 * \f[ 
 *   \begin{bmatrix} \boldsymbol{A} & \boldsymbol{B}_1 \\ \boldsymbol{B}_2 & \boldsymbol{C} \end{bmatrix}. 
 * \f]
 */
template <typename scalar>
class Matrices
{
    /*
    2x2 block matrix:
    ----------------

    *-----*-----*
    |  A  | B_1 |
    *-----*-----*
    | B_2 |  C  |
    *-----*-----*

    B includes the values of B_2 if the contact is fully closed.
    B_T includes the values of B_1 if the contact is fully closed.
    B_G includes the values of B with the M matrix included regardless if we have a Signorini problem or not.
    */
public:
    //
    // Convenience typedefs
    //
    typedef Map::local_ordinal_type local_ordinal_type;
    typedef Map::global_ordinal_type global_ordinal_type;
    typedef Map::node_type node_type;
    typedef Map::tpetra_map_type tpetra_map_type;
    typedef Map::xpetra_bmap_type xpetra_bmap_type;
    typedef Tpetra::CrsGraph<local_ordinal_type, global_ordinal_type, node_type> tpetra_graph_type;
    typedef Tpetra::CrsMatrix<scalar, local_ordinal_type, global_ordinal_type, node_type> tpetra_crs_type;

    typedef Xpetra::Matrix<scalar, local_ordinal_type, global_ordinal_type, node_type> xpetra_mat_type;
    typedef Xpetra::CrsMatrix<scalar, local_ordinal_type, global_ordinal_type, node_type> xpetra_crs_type;
    typedef Xpetra::CrsMatrixWrap<scalar, local_ordinal_type, global_ordinal_type, node_type> xpetra_wcrs_type;
    typedef Xpetra::BlockedCrsMatrix<scalar, local_ordinal_type, global_ordinal_type, node_type> xpetra_bcrs_type;
    typedef Xpetra::TpetraCrsMatrix<scalar, local_ordinal_type, global_ordinal_type, node_type> xpetra_tcrs_type;

    Teuchos::RCP<tpetra_crs_type> K;
    Teuchos::RCP<tpetra_crs_type> S;
    Teuchos::RCP<tpetra_crs_type> L;
    Teuchos::RCP<tpetra_crs_type> B;
    Teuchos::RCP<tpetra_crs_type> B_T;
    Teuchos::RCP<tpetra_crs_type> B_1;
    Teuchos::RCP<tpetra_crs_type> B_2;
    Teuchos::RCP<tpetra_crs_type> B_G;
    Teuchos::RCP<tpetra_crs_type> C;
    Teuchos::RCP<tpetra_crs_type> Cb;

    Teuchos::RCP<xpetra_crs_type> xA;
    Teuchos::RCP<xpetra_crs_type> xB;
    Teuchos::RCP<xpetra_crs_type> xB_T;
    Teuchos::RCP<xpetra_crs_type> xC;

    Teuchos::RCP<xpetra_wcrs_type> xwA;
    Teuchos::RCP<xpetra_wcrs_type> xwB;
    Teuchos::RCP<xpetra_wcrs_type> xwB_T;
    Teuchos::RCP<xpetra_wcrs_type> xwC;

    Teuchos::RCP<xpetra_bcrs_type> blockedMatrix;

    Teuchos::RCP<xpetra_mat_type> A;

    Matrices(Teuchos::RCP<Graph> graph);
};

/**
 * @brief Matrices constructor
 * 
 * This constructor allocates the memory for the matrix \f$\boldsymbol{A}\f$.
 * The other matrices and the filling of the entries are done later.
 * 
 * Argument:
 * <ul>
 *  <li> graph: an RCP to a Graph object.
 * </ul>
 */
template <typename scalar>
Matrices<scalar>::Matrices(Teuchos::RCP<Graph> graph)
{
    if (graph->K != Teuchos::null)
    {
        K = Teuchos::rcp(new tpetra_crs_type(graph->K));
        K->fillComplete();
    }
    if (graph->S != Teuchos::null)
    {
        S = Teuchos::rcp(new tpetra_crs_type(graph->S));
        S->fillComplete(graph->S->getDomainMap(), graph->S->getRangeMap());
    }
    if (graph->L != Teuchos::null)
    {
        L = Teuchos::rcp(new tpetra_crs_type(graph->L));
        L->fillComplete();
    }
}
}; // namespace katropton

#endif //KATROPTON_MATRICES_H
