#ifndef KATROPTON_WEIGHTSLIST_H
#define KATROPTON_WEIGHTSLIST_H

#include "katropton.h"

#include "Map.h"
#include "wWeight.h"
#include "wTag.h"

#include <Tpetra_Map.hpp>
#include <Tpetra_Vector.hpp>
#include <Teuchos_RCP.hpp>
#include <Kokkos_ViewFactory.hpp>

#include "EnsembleTraits.h"

using Teuchos::RCP;

namespace katropton
{

/**
 * @brief Class which is used to store the list of the weights used
 * for the residual computation.
 */
template <typename scalar>
class WeightsList
{
    typedef Map::local_ordinal_type local_ordinal_type;
    typedef Map::global_ordinal_type global_ordinal_type;

private:
    Kokkos::View<int **, Kokkos::LayoutRight> weights_nodesList;
    Kokkos::View<scalar **, Kokkos::LayoutRight> weights_values_list;

    local_ordinal_type weights_number;
    local_ordinal_type weights_size;
    local_ordinal_type weights_nodes;

    local_ordinal_type weights_dofs;
    local_ordinal_type weights_values;

public:
    WeightsList(Problem &pbl, Teuchos::RCP<Map> map, Teuchos::RCP<ElementsList> elementsList);

    /**
     * @brief Return the number of weighted regions.
     */
    inline local_ordinal_type getWeightRegionsNumber()
    {
        return weights_number;
    }

    /**
     * @brief Return the number of nodes for a
     * given weighted region.
     * 
     * Argument:
     * <ul>
     *  <li> i: the weighted region ID.
     * </ul>
     */
    inline local_ordinal_type getWeightRegionSize(local_ordinal_type i)
    {
        return weights_nodesList(i, weights_size);
    }

    /**
     * @brief Return the global ID of the node n1 of 
     * a given weighted region.
     * 
     * Arguments:
     * <ul>
     *  <li> i: the weighted region ID,
     *  <li> n1: the ID of the node on the current BC.
     * </ul>
     */
    inline local_ordinal_type getNode(local_ordinal_type i, local_ordinal_type n1)
    {
        return weights_nodesList(i, weights_nodes + n1);
    }

    /**
     * @brief Return whether the DOF j is weighted for 
     * a given weighted region.
     * 
     * Arguments:
     * <ul>
     *  <li> i: the weighted region ID,
     *  <li> j: the index of the DOF (0-3 if thermomechanical problems).
     * </ul> 
     */
    inline bool getWeightDof(local_ordinal_type i, local_ordinal_type j)
    {
        return (weights_values_list(i, weights_dofs + j) == scalar(1.));
    }

    /**
     * @brief Return the value of weight for the
     * DOF for a given weighted region.
     * 
     * Arguments:
     * <ul>
     *  <li> i: the weighted region ID,
     *  <li> j: the index of the DOF (0-3 if thermomechanical problems).
     * </ul> 
     */
    inline scalar getWeightValue(local_ordinal_type i, local_ordinal_type j)
    {
        return weights_values_list(i, weights_values + j);
    }
};

/**
 * @brief WeightsList constructor
 *
 * This constructor allocates 2 Kokkos::View which store all the required information related to each weighted region, including:
 * <ul>
 *  <li> all the weighted nodes,
 *  <li> all the values of the weights.
 * </ul>
 *
 * Arguments:
 * <ul>
 *  <li> pbl: a Problem object,
 *  <li> map: an RCP to a Map object,
 *  <li> elementsList: an RCP to an ElementList object.
 * </ul>
 */
template <typename scalar>
WeightsList<scalar>::WeightsList(Problem &pbl, Teuchos::RCP<Map> map, Teuchos::RCP<ElementsList> elementsList)
{
    typedef EnsembleTraits<scalar> ET;
    const int ensemble_size = ET::size;

    local_ordinal_type numDPN = map->numPrimalDPN;

    weights_number = pbl.Weights.size();
    local_ordinal_type maxNumNodesPerWeight = 0;

    std::vector<std::vector<global_ordinal_type>> WeightNodes = {};

    for (auto i = 0; i < weights_number; ++i)
    {
        std::vector<global_ordinal_type> tmpMyWeightNodes = {};
        std::vector<global_ordinal_type> myWeightNodes = {};

        for (auto j = 0; j < pbl.Weights[i]->nodes.size(); ++j)
            tmpMyWeightNodes.push_back(pbl.Weights[i]->nodes[j]->row);

        std::sort(tmpMyWeightNodes.begin(), tmpMyWeightNodes.end());

        if (tmpMyWeightNodes.size() >= 1)
            myWeightNodes.push_back(tmpMyWeightNodes[0]);

        for (auto ii = 1; ii < tmpMyWeightNodes.size(); ++ii)
            if (tmpMyWeightNodes[ii] != tmpMyWeightNodes[ii - 1])
                myWeightNodes.push_back(tmpMyWeightNodes[ii]);

        WeightNodes.push_back(myWeightNodes);
        if (maxNumNodesPerWeight < myWeightNodes.size())
            maxNumNodesPerWeight = myWeightNodes.size();
    }

    weights_size = 0;
    weights_nodes = 1;
    weights_nodesList = Kokkos::View<int **, Kokkos::LayoutRight>("R", weights_number, weights_nodes + maxNumNodesPerWeight);

    for (auto i = 0; i < weights_number; ++i)
    {
        weights_nodesList(i, weights_size) = WeightNodes[i].size();

        for (auto j = 0; j < WeightNodes[i].size(); ++j)
            weights_nodesList(i, weights_nodes + j) = WeightNodes[i][j];
    }

    weights_dofs = 0;
    weights_values = weights_dofs + numDPN;
    weights_values_list = Kokkos::View<scalar **, Kokkos::LayoutRight>("R", weights_number, 2 * numDPN);
    for (auto i = 0; i < weights_number; ++i)
    {
        for (auto j = 0; j < numDPN; ++j)
            if (pbl.Weights[i]->which_dof[j])
                weights_values_list(i, weights_dofs + j) = 1;
            else
                weights_values_list(i, weights_dofs + j) = 0;

        for (int s = 0; s < ensemble_size; ++s)
        {
            ET::coeff(weights_values_list(i, weights_values + 0), s) = pbl.Weights[i]->x_values[s];
            ET::coeff(weights_values_list(i, weights_values + 1), s) = pbl.Weights[i]->y_values[s];
            ET::coeff(weights_values_list(i, weights_values + 2), s) = pbl.Weights[i]->z_values[s];
            if (numDPN == 4)
                ET::coeff(weights_values_list(i, weights_values + 3), s) = pbl.Weights[i]->T_values[s];
        }
    }
}
}; // namespace katropton

#endif //KATROPTON_WEIGHTSLIST_H