#!/usr/bin/env python
# -*- coding: latin-1; -*-

from __future__ import print_function
from __future__ import division
from builtins import range
from past.utils import old_div
from scipy.sparse.linalg import spsolve
from scipy.sparse.linalg import eigs
from scipy.linalg import eig
from scipy.sparse import spdiags
import numpy as np
import matplotlib.pyplot as plt
import katropton.convergence as c


def analytical_eigenvectors(n, kappa):
    values = np.zeros((n, ))
    vecs = np.zeros((n, n))

    h = 1. / (n + 1)

    k = np.arange(1, n + 1, 1)
    values = 2 * kappa * (1 - np.cos(k * h * np.pi))

    for i in range(0, n):
        vecs[i, :] = np.sin((i + 1) * k * h * np.pi)

    for i in range(0, n):
        norm = np.linalg.norm(vecs[:, i])
        vecs[:, i] /= norm
    return values, vecs


def omegas_lambdas_nf(n, nkappa, nf):
    N = len(nkappa)
    norm = 0.
    for i in range(0, N):
        norm += np.linalg.norm(nf[i])**2
    norm = np.sqrt(norm)
    lambdas_all = np.array([])
    omegas_all = np.array([])
    for i in range(0, N):
        nf_tmp = old_div(nf[i], norm)
        lambdas, vecs = analytical_eigenvectors(n, nkappa[i])
        c = np.dot(vecs.T, nf_tmp)

        omegas = np.zeros((len(c), ))
        for i in range(0, len(omegas)):
            omegas[i] = np.linalg.norm(c[i])**2

        lambdas_all = np.append(lambdas_all, lambdas)
        omegas_all = np.append(omegas_all, omegas)

    indices = np.argsort(lambdas_all)
    lambdas = lambdas_all[indices]
    omegas = omegas_all[indices]

    return lambdas, omegas


def omegas_lambdas_f(n, kappa, f):
    lambdas, vecs = analytical_eigenvectors(n, kappa)
    norm = np.linalg.norm(f)
    f_tmp = old_div(f, norm)
    c = np.dot(vecs.T, f_tmp)

    omegas = np.zeros((len(c), ))
    for i in range(0, len(omegas)):
        omegas[i] = np.linalg.norm(c[i])**2

    return lambdas, omegas


def write_on_disk(name, res):
    i_max = len(res)
    for i in range(0, i_max):
        if res[i] == 0.:
            i_max = i + 1
            break
    tmp = np.zeros((i_max, 2))
    tmp[:, 1] = res[0:i_max]
    tmp[:, 0] = np.arange(0, i_max)
    np.savetxt(name, tmp)


def main():
    n = 6
    kappa_1 = 1.
    kappa_2 = 1.5

    f_1 = np.zeros((n, ))
    f_2 = np.zeros((n, ))

    f_1[3] = 1.
    f_2[-1] = 1.

    print(f_1)
    print(f_2)

    lambdas_11, omegas_11 = omegas_lambdas_f(n, kappa_1, f_1)
    lambdas_12, omegas_12 = omegas_lambdas_f(n, kappa_1, f_2)

    lambdas_21, omegas_21 = omegas_lambdas_f(n, kappa_2, f_1)
    lambdas_22, omegas_22 = omegas_lambdas_f(n, kappa_2, f_2)

    print(np.round(lambdas_11, 3))
    print(np.round(omegas_11, 3))

    print(np.round(lambdas_12, 3))
    print(np.round(omegas_12, 3))

    print(np.round(lambdas_21, 3))
    print(np.round(omegas_21, 3))

    print(np.round(lambdas_22, 3))
    print(np.round(omegas_22, 3))

    residual_11 = c.compute_convergence_serial_normal(20, n, omegas_11,
                                                      lambdas_11)
    residual_12 = c.compute_convergence_serial_normal(20, n, omegas_12,
                                                      lambdas_12)
    residual_21 = c.compute_convergence_serial_normal(20, n, omegas_21,
                                                      lambdas_21)
    residual_22 = c.compute_convergence_serial_normal(20, n, omegas_22,
                                                      lambdas_22)

    lambdas_1, omegas_1 = omegas_lambdas_nf(n, [kappa_1, kappa_1], [f_1, f_2])
    lambdas_2, omegas_2 = omegas_lambdas_nf(n, [kappa_2, kappa_2], [f_1, f_2])
    lambdas_3, omegas_3 = omegas_lambdas_nf(n, [kappa_1, kappa_2], [f_1, f_1])
    lambdas_4, omegas_4 = omegas_lambdas_nf(n, [kappa_1, kappa_2], [f_2, f_2])
    lambdas_5, omegas_5 = omegas_lambdas_nf(n, [kappa_1, kappa_2], [f_1, f_2])

    residual_1 = c.compute_convergence_serial_normal(20, 2 * n, omegas_1,
                                                     lambdas_1)
    residual_2 = c.compute_convergence_serial_normal(20, 2 * n, omegas_2,
                                                     lambdas_2)
    residual_3 = c.compute_convergence_serial_normal(20, 2 * n, omegas_3,
                                                     lambdas_3)
    residual_4 = c.compute_convergence_serial_normal(20, 2 * n, omegas_4,
                                                     lambdas_4)
    residual_5 = c.compute_convergence_serial_normal(20, 2 * n, omegas_5,
                                                     lambdas_5)

    write_on_disk('res_11.txt', residual_11)
    write_on_disk('res_12.txt', residual_12)
    write_on_disk('res_21.txt', residual_21)
    write_on_disk('res_22.txt', residual_22)
    write_on_disk('res_1.txt', residual_1)
    write_on_disk('res_2.txt', residual_2)
    write_on_disk('res_3.txt', residual_3)
    write_on_disk('res_4.txt', residual_4)
    write_on_disk('res_5.txt', residual_5)

    print(np.round(lambdas_1, 3))
    print(np.round(omegas_1, 3))
    print(np.round(lambdas_2, 3))
    print(np.round(omegas_2, 3))
    print(np.round(lambdas_3, 3))
    print(np.round(omegas_3, 3))
    print(np.round(lambdas_4, 3))
    print(np.round(omegas_4, 3))
    print(np.round(lambdas_5, 3))
    print(np.round(omegas_5, 3))


if __name__ == "__main__":
    main()
