/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wFct0.h"
#include "wElement.h"
#include "wNode.h"
#include "wCache.h"

using namespace tbox;

void PwLf::add(double x, double y)
{
    auto it = pts.find(x);
    if (it != pts.end())
    {
        std::ostringstream str;
        str << "PwLf: value at " << x << " already exist!\n";
        throw std::runtime_error(str.str());
    }
    pts[x] = y;
}

double PwLf::eval(double x) const
{
    auto it1 = pts.begin();
    if (it1 == pts.end())
        throw std::runtime_error("PwLf is empty!");
    auto it2 = it1;
    ++it2;
    if (it2 == pts.end())
        return it1->second; //fct is a constant
    auto it3 = it2;
    ++it3;

    for (; it3 != pts.end(); ++it1, ++it2, ++it3)
        if (x <= it2->first)
            break;

    // eval using it1-it2
    double x1 = it1->first;
    double y1 = it1->second;
    double x2 = it2->first;
    double y2 = it2->second;
    return y1 + (y2 - y1) / (x2 - x1) * (x - x1);
}

void PwLf::write(std::ostream &out) const
{
    out << "PwLf=";
    for (auto &p : pts)
        out << "\tx=" << p.first << "  y=" << p.second << '\n';
}

//-------------------------------------------------------------------------

double Fct0::evalD(Element &e, std::vector<double> const &u, size_t k) const
{
    throw std::runtime_error("Fct0::evalD not implemented\n");
}

//-------------------------------------------------------------------------

void Fct0C::write(std::ostream &out) const
{
    out << "Fct0C=" << v;
}

//-------------------------------------------------------------------------

double
Fct0XYZ::eval(Element &e, std::vector<double> const &u, size_t k) const
{
    Cache &cache = e.getVCache();
    Eigen::VectorXd const &ff = cache.getSf(k);

    // position pg en x,y
    Eigen::Vector3d xpg(0, 0, 0);
    for (size_t i = 0; i < e.nodes.size(); ++i)
        xpg += ff(i) * e.nodes[i]->pos;

    return this->eval(xpg);
}

void Fct0XYZ::write(std::ostream &out) const
{
    out << "Fct0XYZ";
}

//-------------------------------------------------------------------------

double
Fct0U::eval(Element &e, std::vector<double> const &u, size_t k) const
{
    Cache &cache = e.getVCache();
    Eigen::VectorXd const &ff = cache.getSf(k);

    // u au pg
    double upg = 0.0;
    for (size_t i = 0; i < e.nodes.size(); ++i)
        upg += ff(i) * u[e.nodes[i]->row];

    return this->eval(upg);
}

void Fct0U::write(std::ostream &out) const
{
    out << "Fct0U";
}
