/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wTri3.h"
#include "wSfTri3.h"
#include "wGaussTri3.h"
#include "wCacheTri3.h"
#include "wMemTri3.h"
#include "wNode.h"
#include "wFct0.h"
#include "wFct1.h"
#include "wFct2.h"

using namespace tbox;

Tri3::Tri3(int n, Tag *_ptag, Tag *_etag,
           std::vector<Tag *> &_parts, std::vector<Node *> &nods) : Element(n, _ptag, _etag, _parts, nods)
{
    pmem = new MemTri3(*this, static_cast<int>(getCache().gauss.getN()));
}
Tri3::~Tri3()
{
    if (pmem)
        delete pmem;
}

/**
 * @brief Return private memory handling Jacobian for this Tri3 element 
 */
Mem &Tri3::getVMem()
{
    return getMem();
}

/**
 * @brief Return cache holding shape functions and Gauss points for all Tri3 elements
 */
Cache &Tri3::getVCache()
{
    return getCache();
}

/**
 * @brief Compute element unit normal vector
 */
Eigen::Vector3d Tri3::normal()
{
    Eigen::Vector3d x1 = nodes[0]->pos;
    Eigen::Vector3d x2 = nodes[1]->pos;
    Eigen::Vector3d x3 = nodes[2]->pos;

    return ((x2 - x1).cross(x3 - x1)).normalized();
}

/**
 * @brief Compute Jacobian determinant at Gauss point k
 * 
 * detJ = norm(x_i dN_i,xi X x_i dN_i,eta)
 */
double Tri3::buildJ(size_t k) const
{
    Eigen::MatrixXd const &dff = getCache().getDsf(k);

    Eigen::Vector3d t0 = Eigen::Vector3d::Zero();
    Eigen::Vector3d t1 = Eigen::Vector3d::Zero();
    size_t i = 0;
    for (auto it = nodes.begin(); it != nodes.end(); ++it, ++i)
    {
        t0 += (*it)->pos * dff(0, i);
        t1 += (*it)->pos * dff(1, i);
    }
    return t0.cross(t1).norm();
}
/**
 * @brief Compute Jacobian matrix and determinant at Gauss point k
 * 
 * J_ij(2,2) = diN_n * xj_n (n is the node)
 */
double Tri3::buildJ(size_t k, Eigen::MatrixXd &J) const
{
    Eigen::MatrixXd const &dff = getCache().getDsf(k);

    J = Eigen::Matrix2d::Zero();
    size_t i = 0;
    for (auto it = nodes.begin(); it != nodes.end(); ++it, ++i)
        J += dff.col(i) * (*it)->pos.transpose().block<1, 2>(0, 0);
    return J.determinant();
}

/**
 * @brief Compute stiffness matrix for scalar field
 * 
 * K_ij(3,3) = int f*dN_j*dN_i dV
 */
Eigen::MatrixXd Tri3::buildK(std::vector<double> const &u, Fct0 const &fct)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    // Gauss integration
    Eigen::MatrixXd K = Eigen::Matrix3d::Zero();
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Jacobian inverse and shape functions
        Eigen::Matrix2d const &J = mem.getJinv(k);
        Eigen::MatrixXd const &dff = cache.getDsf(k);

        // Elementary stiffness matrix
        K += (J * dff).transpose() * J * dff * fct.eval(*this, u, k) * gauss.getW(k) * mem.getDetJ(k);
    }
    return K;
}

/**
 * @brief Compute stiffness matrix for scalar field
 * 
 * K_ij(3,3) = int [f]*dN_j*dN_i dV
 */
Eigen::MatrixXd Tri3::buildK(std::vector<double> const &u, Fct2 const &fct, bool fake)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;

    Eigen::MatrixXd K = Eigen::Matrix3d::Zero();
    if (fake) // fake run - fill MPI job list
    {
        for (size_t k = 0; k < gauss.getN(); ++k)
        {
            Eigen::MatrixXd fk;
            fct.eval(*this, u, k, fk, fake);
        }
    }
    else // true run - use MPI cached results
    {
        // Gauss integration
        MemTri3 &mem = getMem();
        for (size_t k = 0; k < gauss.getN(); ++k)
        {
            // Jacobian inverse, shape functions and function evaluation
            Eigen::Matrix2d const &J = mem.getJinv(k);
            Eigen::MatrixXd const &dff = cache.getDsf(k);
            Eigen::MatrixXd fk;
            fct.eval(*this, u, k, fk, fake);

            // Elementary stiffness matrix
            K += (fk * J * dff).transpose() * J * dff * gauss.getW(k) * mem.getDetJ(k);
        }
    }
    return K;
}

/**
 * @brief Compute mechanical stiffness matrix
 * 
 * K_ij(6,6) = int eps_k H_ijkl eps_l dV = B(6,3)H(3,3)B(3,6)
 */
Eigen::MatrixXd Tri3::buildK_mechanical(Eigen::MatrixXd const &H)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    // Gauss integration
    Eigen::MatrixXd K = Eigen::MatrixXd::Zero(6, 6);
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Jacobian inverse
        Eigen::Matrix2d const &invJ = mem.getJinv(k);
        // Fill B matrix (shape functions)
        Eigen::MatrixXd const &dff = cache.getDsf(k);
        Eigen::MatrixXd B = Eigen::MatrixXd::Zero(3, 6);
        for (size_t i = 0; i < nodes.size(); ++i)
        {
            // Compute [Ni,x Ni,y]
            Eigen::Vector2d dN = invJ * dff.col(i);
            B(0, 2 * i) = dN(0);
            B(1, 2 * i + 1) = dN(1);
            B(2, 2 * i + 1) = dN(0);
            B(2, 2 * i) = dN(1);
        }

        // Compute stiffness matrix
        K += B.transpose() * H * B * gauss.getW(k) * mem.getDetJ(k);
    }
    return K;
}

/**
 * @brief Compute (first part of) derivative stiffness matrix for scalar field (Newton method)
 * 
 * K_ij(3,3) = int (df du*dN_j du*dN_i) dV
 */
Eigen::MatrixXd Tri3::buildDK(std::vector<double> const &u, Fct0 const &fct)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    // Gauss integration
    Eigen::MatrixXd K = Eigen::Matrix3d::Zero();
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Jacobian inverse, shape functions and gradient
        Eigen::Matrix2d const &J = mem.getJinv(k);
        Eigen::MatrixXd const &dff = cache.getDsf(k);
        Eigen::VectorXd gradu = computeGrad(u, k);

        // Elementary stiffness matrix
        K += (gradu.transpose() * J * dff).transpose() * gradu.transpose() * J * dff * fct.evalD(*this, u, k) * gauss.getW(k) * mem.getDetJ(k);
    }
    return K;
}

/**
 * @brief Compute flux vector for a scalar field
 * 
 * s_i(3) = int f*N_i dV
 */
Eigen::VectorXd Tri3::builds(std::vector<double> const &u, Fct0 const &f)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    // Gauss integration
    Eigen::VectorXd s = Eigen::Vector3d::Zero();
    for (size_t k = 0; k < gauss.getN(); ++k)
        s += cache.getSf(k) * f.eval(*this, u, k) * gauss.getW(k) * mem.getDetJ(k);
    return s;
}

/**
 * @brief Compute flux vector for scalar field
 * 
 * s_i(3) = int f_j*n_j * N_i dV 
 */
Eigen::VectorXd Tri3::builds(std::vector<double> const &u, Fct1 const &f)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    // Gauss integration
    Eigen::VectorXd s = Eigen::Vector3d::Zero();
    for (size_t k = 0; k < gauss.getN(); ++k)
        s += cache.getSf(k) * (f.eval(*this, u, k).dot(normal())) * gauss.getW(k) * mem.getDetJ(k);
    return s;
}

/**
 * @brief Compute derivative of flux vector for a scalar field (Newton method)
 * 
 * s_i(3) = int f*du*dN_i dV
 */
Eigen::VectorXd Tri3::buildDs(std::vector<double> const &u, Fct0 const &f)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    // Gauss integration
    Eigen::VectorXd s = Eigen::Vector3d::Zero();
    for (size_t k = 0; k < gauss.getN(); ++k)
        s += computeGrad(u, k).transpose() * mem.getJinv(k) * cache.getDsf(k) * f.eval(*this, u, k) * gauss.getW(k) * mem.getDetJ(k); // elementary flux vector
    return s;
}

/**
 * @brief Compute mass matrix
 * 
 * S_ij(3,3) = int N_i*N_j dV
 */
Eigen::MatrixXd Tri3::buildS()
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    // Gauss integration
    Eigen::MatrixXd S = Eigen::Matrix3d::Zero();
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Shape functions
        Eigen::VectorXd const &ff = cache.getSf(k);

        // Mass matrix
        S += ff * ff.transpose() * gauss.getW(k) * mem.getDetJ(k);
    }
    return S;
}

/**
 * @brief Compute gradient of a scalar field at Gauss point k
 */
Eigen::VectorXd Tri3::computeGrad(std::vector<double> const &u, size_t k)
{
    // Solution vector
    Eigen::Vector3d ue;
    for (size_t i = 0; i < nodes.size(); ++i)
        ue(i) = u[nodes[i]->row];
    // Gradient in global axis: inv(J)_ij * d_jN_i * ue_i
    return getMem().getJinv(k) * getCache().getDsf(k) * ue;
}

/**
 * @brief Compute square of gradient of a scalar field at Gauss point k
 */
double Tri3::computeGrad2(std::vector<double> const &u, size_t k)
{
    return computeGrad(u, k).squaredNorm();
}

/**
 * @brief Compute flux
 * 
 * qV_i(3) = int [f] gradu N_i dV
 */
Eigen::VectorXd Tri3::computeqint(std::vector<double> const &u, Fct2 const &fct)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    // Gauss integration
    Eigen::VectorXd qV = Eigen::Vector3d::Zero();
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Shape functions, gradient and flux evaluation
        Eigen::MatrixXd const &dff = cache.getDsf(k);
        Eigen::MatrixXd fk(2, 2);
        fct.eval(*this, u, k, fk, false);

        // Elementary flux vector
        qV += (fk * computeGrad(u, k)).transpose() * mem.getJinv(k) * dff * gauss.getW(k) * mem.getDetJ(k);
    }
    return qV;
}

/**
 * @brief Compute flux
 * 
 * qV_i(3) = int -[f] gradu dV
 */
Eigen::VectorXd Tri3::computeqV(std::vector<double> const &u, Fct2 const &fct)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    // Gauss integration
    Eigen::VectorXd qV = Eigen::Vector2d::Zero();
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Gradient and matrix function evaluation
        Eigen::VectorXd gradk = computeGrad(u, k);
        Eigen::MatrixXd fk(2, 2);
        fct.eval(*this, u, k, fk, false);

        qV -= fk * gradk * gauss.getW(k) * mem.getDetJ(k);
    }
    return qV;
}

/**
 * @brief Compute intergal of scalar function
 * 
 * I = int f dV
 */
double Tri3::computeV(std::vector<double> const &u, Fct0 const &fct)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    // Gauss integration
    double V = 0.0;
    for (size_t k = 0; k < gauss.getN(); ++k)
        V += fct.eval(*this, u, k) * gauss.getW(k) * mem.getDetJ(k);
    return V;
}

/**
 * @brief Compute intergal of matrix function
 * 
 * I = int [f] dV
 */
Eigen::MatrixXd Tri3::computeV(std::vector<double> const &u, Fct2 const &fct)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    // Gauss integration
    Eigen::MatrixXd out = Eigen::MatrixXd::Zero(2, 2); /* @todo out should be resized to match fk size */
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Function evaluation
        Eigen::MatrixXd fk;
        fct.eval(*this, u, k, fk, false);

        out += fk * gauss.getW(k) * mem.getDetJ(k);
    }
    return out;
}

/**
 * @brief Return the element surface
 */
double Tri3::computeV()
{
    return getMem().getVol();
}
