#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Simple 2D wave-propagation test in 1 material

from __future__ import division
from past.utils import old_div
import waves as w
import tbox
import tbox.gmsh as gmsh
import waves.srcs as srcs
import waves.utils as u


def model(p={}):
    # configure/build mesh
    pars = {'Lx': p['Lx'], 'Ly': p['Ly'], 'Nx': p['Nx'], 'Ny': p['Ny']}
    msh = gmsh.MeshLoader("cube.geo", __file__).execute(**pars)
    gmshWriter = tbox.GmshExport(msh)
    pbl = w.Problem(msh)

    # compute time-step size
    Lex = old_div(p['Lx'], p['Nx'])
    Ley = old_div(p['Ly'], p['Ny'])
    dt = min(Lex, Ley) / p['c'] / 10 * 2

    pbl.add(w.Medium(msh, "Volume", p['c']))

    if p['bc'] == 'sinsrc':  # source dans le coin inf gauche
        pbl.add(srcs.SinSource(msh, "Line1", p['freq']))
        pbl.add(w.Boundary(msh, "FaceX1", p['c']))
        pbl.add(w.Boundary(msh, "FaceY1", p['c']))

    elif p['bc'] == 'planewave':  # onde plane de bas en haut
        pbl.add(srcs.SinSource(msh, "FaceX0", p['freq']))
        pbl.add(w.Boundary(msh, "FaceX1", p['c']))

    elif p['bc'] == 'interf1':  # interférence de 2 srcs sur la frontière basse
        s = srcs.MultiSource(msh, "FaceY0")
        s.addSource(tbox.Vector3d(3, 0, 0), old_div(p['freq'], 2))
        s.addSource(tbox.Vector3d(7, 0, 0), p['freq'])
        pbl.add(s)
        pbl.add(w.Boundary(msh, "FaceX0", p['c']))
        pbl.add(w.Boundary(msh, "FaceX1", p['c']))
        pbl.add(w.Boundary(msh, "FaceY1", p['c']))

    elif p['bc'] == 'pulse':  # un pulse qui part du coin inf gauche
        pbl.add(
            srcs.PulseSource(msh, "Line1", p['freq'], shift=0.3, width=100.0))
        pbl.add(w.Boundary(msh, "FaceX1", p['c']))
        pbl.add(w.Boundary(msh, "FaceY1", p['c']))

    elif p['bc'] == 'interf2':  # 2 sources non synchrones au centre du domaine
        s = srcs.MultiSource(msh, "Volume")
        s.addSource(tbox.Vector3d(3 * p['Lx'] / 10, old_div(p['Ly'], 2), 0), p['freq'])
        s.addSource(
            tbox.Vector3d(7 * p['Lx'] / 10, old_div(p['Ly'], 2), 0), 1.05 * p['freq'])
        #pbl.add(s.__disown__())
        pbl.add(s)
        pbl.add(w.Boundary(msh, "FaceX0", p['c']))
        pbl.add(w.Boundary(msh, "FaceX1", p['c']))
        pbl.add(w.Boundary(msh, "FaceY0", p['c']))
        pbl.add(w.Boundary(msh, "FaceY1", p['c']))
    else:
        raise Exception('Unknown BC type (%s)' % p['bc'])

    scheme = w.ForwardEuler(pbl)
    #scheme.wgmsh = w.Wgmsh_WG_ALL # write .pos files
    scheme.ttot = p['ttot']
    scheme.dt = dt
    scheme.savefreq = 5
    return scheme, gmshWriter


def runmodel(**d):
    p = {}
    p['bc'] = 'interf2'
    p['ttot'] = 1.0
    p['Lx'] = 10.0
    p['Ly'] = 10.0
    p['Nx'] = 50
    p['Ny'] = 50
    p['c'] = 10.0
    p['freq'] = 10.0
    p.update(d)

    scheme, writer = model(p)
    u.integrate(scheme, writer)


if __name__ == "__main__":
    runmodel()
